import unittest
import os

class TestDataModality(unittest.TestCase):


    def test_string_pair(self):
        from h5record.dataset import H5Dataset
        from h5record.attributes import String, Integer
        schema = {
            'sentence1': String(name='sentence1'),
            'sentence2': String(name='sentence2'),
            'label': Integer(name='label')
        }

        data = [
            ['HDF5 supports chunking and compression.', 'You may want to experiment', 0],
            ['Starting to load file into an HDF5 file with chunk size','and compression is gzip', 1 ],
            ['Reading from an HDF5 file which you will probably be','about to overwrite! Override this error only if you know what youre doing ', 1],
        ]

        def pair_iter():
            for row in data:
                yield {
                    'sentence1': row[0],
                    'sentence2': row[1],
                    'label': row[2]
                }
        if os.path.exists('question_pair.h5'):
            os.remove('question_pair.h5')

        dataset = H5Dataset(schema, './question_pair.h5', pair_iter())
        for idx in range(len(data)):
            row = dataset[idx]
            sent1, sent2, label = data[idx]
            assert sent1 == row['sentence1']
            assert sent2 == row['sentence2']
            assert label == row['label']

        os.remove('question_pair.h5')

        dataset = H5Dataset(schema, './question_pair.h5', pair_iter(),
            data_length=len(data), chunk_size=4)

        for idx in range(len(data)):
            row = dataset[idx]
            sent1, sent2, label = data[idx]
            assert sent1 == row['sentence1']
            assert sent2 == row['sentence2']
            assert label == row['label']

        os.remove('question_pair.h5')


    def test_unicode(self):
        from h5record.dataset import H5Dataset
        from h5record.attributes import String, Integer
        schema = {
            'sentence1': String(name='sentence1'),
            'sentence2': String(name='sentence2'),
            'label': Integer(name='label')
        }

        data = [
            [  '小明去學校', '結果已經下課了', 0]
        ]

        def pair_iter():
            for row in data:
                yield {
                    'sentence1': row[0],
                    'sentence2': row[1],
                    'label': row[2]
                }
        if os.path.exists('question_pair.h5'):
            os.remove('question_pair.h5')

        dataset = H5Dataset(schema, './question_pair.h5', pair_iter())
        for idx in range(len(data)):
            row = dataset[idx]
            sent1, sent2, label = data[idx]

            assert sent1 == row['sentence1']
            assert sent2 == row['sentence2']
            assert label == row['label']

        os.remove('question_pair.h5')


    def test_image_string_pair(self):
        from h5record.dataset import H5Dataset
        from h5record.attributes import String, Image

        schema = {
            'image': Image(name='image', h=32, w=32),
            'caption': String(name='caption')
        }

        captions = [
            'Lenna profile',
            'Lenna back patch',
            'Lenna lower patch',
            'meme image', 
            'greyscale image'
        ]

        image_paths = [
            'test/images/1.jpeg',
            'test/images/2.jpeg',
            'test/images/3.jpeg',
            'test/images/4.jpeg',
            'test/images/5.jpeg',
        ]
        def pair_iter():
            for (caption, img_path) in zip(captions, image_paths):
                yield {
                    'image': schema['image'].read_image(img_path),
                    'caption': caption
                }
        if os.path.exists('img_caption.h5'):
            os.remove('img_caption.h5')

        dataset = H5Dataset(schema, './img_caption.h5', pair_iter(),
            data_length=len(image_paths), chunk_size=4)

        for idx in range(len(image_paths)):
            data = dataset[idx]
            assert data['caption'] == captions[idx]

        os.remove('img_caption.h5')


        dataset = H5Dataset(schema, './img_caption.h5', pair_iter(),
            compression='lzf')

        for idx in range(len(image_paths)):
            data = dataset[idx]
            assert data['caption'] == captions[idx]

        os.remove('img_caption.h5')



import h5py as h5
import numpy as np

try:
    from PIL import Image as PImage
except ImportError as e:
    pass

class Attribute():

    def append(self, h5, data):
        raise NotImplementedError

    def transform(self, data):
        raise NotImplementedError

class Integer(Attribute):
    '''
        One dimensional data shape
    '''
    dtype = 'int64'
    def __init__(self, name='label'):
        self.name = name
        self.shape = (None, )
        self.max_shape = (None, )

    def append(self, h5, data):
        h5[self.name].resize( h5[self.name].shape[0]+data.shape[0], axis=0)
        h5[self.name][-data.shape[0]:] = data
        return h5

    def transform(self, data):
        return np.array([data])

class Float(Attribute):
    '''
        One dimensional data shape
    '''
    dtype = 'float16'
    def __init__(self, name='label'):
        self.name = name
        self.shape = (None, )
        self.max_shape = (None, )

    def append(self, h5, data):
        h5[self.name].resize( h5[self.name].shape[0]+data.shape[0], axis=0)
        h5[self.name][-data.shape[0]:] = data
        return h5

    def transform(self, data):
        return np.array([data])


class Image(Attribute):

    dtype = 'uint8'
    def __init__(self, h, w, c=3, name='image'):
        self.c = c
        self.h = h
        self.w = w
        self.name = name

        self.shape = (None, self.c, self.h, self.w)
        self.max_shape = (None, self.c, self.h, self.w)

    def read_image(self, filename, resize=True): 
        # read image by file path and return numpy object
        img = PImage.open(filename)
        if resize:
            img  = img.resize((self.w, self.h))
        np_img = np.array(img)
        np_img = np.array(img)
        if len(np_img.shape) == 2:
            np_img = np.expand_dims(np_img, -1)
            np_img = np.repeat(np_img, 3, axis=-1)

        return np.transpose(np_img, (2, 0, 1))

    def append(self, h5, data):
        h5[self.name].resize( h5[self.name].shape[0]+data.shape[0], axis=0)
        h5[self.name][-data.shape[0]:] = data
        return h5

    def transform(self, data):
        data =  np.array(data)
        if len(data.shape) == 3: 
            # ensure data shape is B x C x H x W
            data = np.expand_dims(data, axis=0)
        return data

class Sequence(Attribute):

    dtype = h5.special_dtype(vlen=np.dtype('int32'))

    def __init__(self, name='sequence', sub_attributes=None):
        self.name = name
        self.shape = (1, 1, )
        self.sub_attributes = sub_attributes
        self.max_shape = (None, 1)


    def append(self, h5, data):
        if isinstance(data, dict):
            for key in self.sub_attributes:
                np_seq = data[key]
                np_seq = np.array([np_seq], dtype=self.dtype)
                h5[self.name + '_'+key].resize( h5[self.name + '_'+key].shape[0]+np_seq.shape[0], axis=0)
                h5[self.name + '_'+key][-np_seq.shape[0]:] = np_seq
        elif isinstance(data, np.ndarray):
            h5[self.name].resize( h5[self.name].shape[0]+np_seq.shape[0], axis=0)
            h5[self.name][-np_seq.shape[0]:] = np_seq
        else:
            raise ValueError("invalid data type: {}".format(type(data)))

        return h5


    def transform(self, data):
        if isinstance(data, dict):
            assert self.sub_attributes is not None, "sub attributes not defined"

            for key in self.sub_attributes:
                np_seq = data[key]
                if len(np_seq.shape) == 2:
                    np_seq = np.array([np_seq], dtype=self.dtype)
                    data[key] = np_seq
            return data
        elif isinstance(data, np.ndarray):
            if len(data.shape) == 2: 
                # make sure its B x 1 x sequence length
                np_seq = np.array([data], dtype=self.dtype)
            return np_seq
        else:
            raise ValueError("invalid data type: {}".format(type(data)))

class String(Attribute):

    encoding = 'utf-8'
    dtype = h5.string_dtype(encoding='utf-8')

    def __init__(self, name='string'):
        self.name = name
        self.max_shape = (None, 1)
        self.shape = None

    def append(self, h5, data):
        buf_size = len(data)
        h5[self.name].resize((h5[self.name].shape[0]+buf_size), axis=0)
        h5[self.name][-buf_size:] = data
        return h5

    def transform(self, data):
        assert isinstance(data, str)
        return [data]




import h5py as h5
import numpy as np
from torch.utils.data.dataset import Dataset
import os
from .attributes import (
    String
)

class AtomicFile:
    '''
        Wrapper file for h5 in case multiprocess writes is needed
    '''
    def __init__(self, path):
        self.fd = os.open(path, os.O_RDONLY)
        self.pos = 0

    def seek(self, pos, whence=0):
        if whence == 0:
            self.pos = pos
        elif whence == 1:
            self.pos += pos
        else:
            self.pos = os.lseek(self.fd, pos, whence)
        return self.pos

    def tell(self):
        return self.pos

    def read(self, size):
        b = os.pread(self.fd, size, self.pos)
        self.pos += len(b)
        return b

class H5Dataset(Dataset):

    def __init__(self, schema, save_filename, data_iter=None,
        data_length=None, chunk_size=300, compression=None, 
        transform=None, append_mode=False, verbose=0):

        '''
        Note: 
            * data length must be known value otherwise chunk size will not be enabled
            * chunk size affects reading speed, usually a size of 100-500 is suitable value
            * compression algorithm affects reading speed, so if storage is not your concern is recommended not to enable
        '''

        # normalized schema design to dictionary
        if isinstance(schema, list) or isinstance(schema, tuple):
            schema = {  s.name: s  for s in schema }
        self.schema = schema
        self.save_filename = save_filename
        self.data_length = data_length # dataset maximum size
        self.transform = transform # transform function before returned by index access
        self.append_mode = append_mode # force append ?

        self.chunk_size = None if data_length is None else chunk_size
        assert compression in [None, 'lzf', 'gzip', 'szip']
        self.compression = compression
        if not os.path.exists(self.save_filename):
            self.preprocess(data_iter)

        self.reader = h5.File(self.save_filename, 'r')
        first_key = list(self.schema.keys())[0]
        self.num_entries = self.reader[first_key].shape[0]-1


    def preprocess(self, data_iter):
        idx = 0
        for data in data_iter:
            if idx == 0:
                with h5.File(self.save_filename, 'w', libver='latest', swmr=True) as fout:
                    fout.swmr_mode = True 
                    for key, value in data.items():
                        attribute = self.schema[key]
                        value = attribute.transform(value)
                        max_shape = list(attribute.max_shape)
                        max_shape[0] = self.data_length
                        max_shape = tuple(max_shape)
                        if not isinstance(attribute, String):
                            shape = value.shape
                        else:
                            shape = (len(value), 1)

                        dset = fout.create_dataset(key,data=value, shape=shape, 
                            maxshape=max_shape, 
                            dtype=attribute.dtype, 
                            compression=self.compression )
            else:
                with h5.File(self.save_filename, 'a', libver='latest', swmr=True) as fout:
                    fout.swmr_mode = True 
                    for key, value in data.items():
                        attribute = self.schema[key]
                        value = attribute.transform(value)
                        attribute.append(fout, value)

            idx += 1

    def __len__(self):
        return self.num_entries


    def __getitem__(self, idx):
        data = {}
        for key in self.schema.keys():
            raw_output = self.reader[key][idx]
            if isinstance(self.schema[key], String):
                data[key] = raw_output[0].decode(self.schema[key].encoding )
            else:
                data[key] = raw_output

        if self.transform is not None:
            return self.transform(data)
        
        return data





'''
setup.py - a setup script
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Authors:
    @theblackcat
'''
from setuptools import setup, Extension

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('LICENSE', 'r') as f:
    license_ = f.read()

with open('README.md', 'r') as f:
    readme = f.read()


test_requirements = [
    'Pillow',
]

setup(
    name='h5record',
    version='1.0.1',
    description='Large data storage for pytorch',
    long_description=readme,
    long_description_content_type="text/markdown",
    author='theblackcat',
    author_email='zhirui09400@gmail.com',
    url='https://github.com/theblackcat102/h5record',
    keywords='data processing',
    packages=['h5record'],
    install_requires=[
        'torch',
        'h5py',
        'numpy'
    ],
    tests_require=test_requirements,
    license='MIT License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)
